/**********************************************************************\
|				   				       |
|                        Temporal Access Method:	               |
|                     	  Overlapping B+-trees                         |
|   		                                        	       |
|   	 	      Last modified: October 8, 2000         	       |
|								       |
\**********************************************************************\

			    COPYRIGHT NOTICE

This source code was created by Mr. Theodoros Tzouramanis
<theo@delab.csd.auth.gr>, from Aristotle University of Thessaloniki,
Department of Informatics, Thessaloniki, Greece,
under the supervision of Prof. Yannis Manolopoulos and
Prof. Nikos Lorentzos.
Permission is granted to copy this software, to redistribute it
on a nonprofit basis, and to use it for any purpose, subject to
the following restrictions and understandings:
a. Any copy made of this software must include this copyright notice
   in full.
b. All materials developed as a consequence of the use of this
   software shall duly acknowledge such use, in accordance with the usual
   standards of acknowledging credit in academic research.
c. The authors have made no warranty or representation that the
   operation of this software will be error-free or suitable for any
   application, and they are under no obligation to provide any
   services, by way of maintenance, update, or otherwise.  The software
   is an experimental prototype offered on an as-is basis.
d. Redistribution for profit requires the express, written permission
   of the authors.


/*~~~~~~~~~~~~~~~~~~~~~~~~~ files to include ~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

#include "overbt.h"
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <alloc.h>
#include <ctype.h>
#include <io.h>
#include <fcntl.h>
//#include <dos.h>

/*-------------------------------------------------------------------------*/

void main()
{

const
 MAXUpdNumKeys = 1000;

FILE 	    *RECfile,    /* T-tree file */
	    *TTTfile,    /* TTT file */
	    *DATAfile,   /* input file */
	    *PRINTfile;  /* report file used only in experimentation */
SL          *firstInREClist = NULL; /* linked list of reocrds for batch insertion, deletion or update */
STACKLINK   *firstTrans = NULL, *lastTrans = NULL, *askedtrans; /* firstTrans and lastTrans are the first and last pointers in the double linked list of transation timestamps */
int 	    trans, lab, flag = FALSE, found;
char        RECfilename[] = "rec.dat", TTTfilename[] = "trans.dat",
	    DATAfilename[] = "input.dat", PRINTfilename[] = "experime.txt",
	    command;

clrscr();

PRINTfile = openWTfile(PRINTfilename);
fseek(PRINTfile, 0L, SEEK_END);
DATAfile  = openRTfile(DATAfilename);

fprintf(PRINTfile, "-----------------MAXex = %d -----------------\n", MAXex);
fprintf(stdout, "Load database from disk (Y/N) : ");
command = getch();
fprintf(stdout,"%c\n",command);

if( tolower(command)=='n' )
 {
 TTTfile = openWfile(TTTfilename);
 RECfile  = openWfile(RECfilename);
 fprintf(stdout, "First transaction.");
 fprintf(stdout, "\n");
 fprintf(PRINTfile, "First transaction.");
 fprintf(PRINTfile, "\n");
 getmemForSTACK(&firstTrans,&lastTrans);
 }
else
 {
 initCounters(&CountTranspos,&CountRecpos);
 TTTfile = openRWfile(TTTfilename);
 RECfile  = openRWfile(RECfilename);
 fprintf(stdout, " Last transaction-no = %d .\n" , CountTranspos);
 fprintf(stdout, "");
 fprintf(PRINTfile, " Last transaction-no = %d .\n" , CountTranspos);
 fprintf(PRINTfile, "");
 initSTACKlist(&firstTrans, &lastTrans, TTTfile);
 }

do
 {
 fprintf(stdout,"\nCOMMAND:(m)odify,(s)earch,(p)rint,(n)ew ");
 fprintf(stdout,"trans,(c)ounters,(q)uit : ");
 command = getch();
 fprintf(stdout,"%c\n",command);
 fprintf(PRINTfile,"\nCOMMAND:(m)odify,(s)earch,(p)rint,(n)ew ");
 fprintf(PRINTfile,"trans,(c)ounters,(q)uit: %c",command);

 switch(tolower(command))
  {
  case 'm' : makelist(&firstInREClist, DATAfile, MAXUpdNumKeys);
	     batch_modify(&firstInREClist, &(lastTrans->mainroot.treepointer), RECfile);
	     printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n", WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	     fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	     break;

  case 's' : search(firstTrans, lastTrans, RECfile, PRINTfile);
	     break;

  case 'p' : fprintf(stdout, "\nPrint which transaction ? ");
	     fscanf(stdin, "%d",&trans);
	     fprintf(stdout, "\nRollback database .\n ");
	     fprintf(stdout, "\nTransaction = %d\n\n",trans);
	     fprintf(PRINTfile, "\nPrint which transaction ? %d",trans);
	     fprintf(PRINTfile, "\nRollback database .\n ");
	     fprintf(PRINTfile, "\nTransaction = %d\n\n",trans);
	     askedtrans = findtrans(firstTrans, trans);

	     if(askedtrans!=NULL)
	      {
	      PrintedKeys = NodesInTree = LeafNodes = 0;
	      printtree(askedtrans->mainroot.treepointer, &lab, RECfile, PRINTfile);
	      fprintf(stdout, "\nKeys = %d ..... nodes = %d..... LeafNodes = %d\n\n",PrintedKeys,NodesInTree,LeafNodes);
	      fprintf(PRINTfile, "\nKeys = %d ..... nodes = %d..... LeafNodes = %d\n\n",PrintedKeys,NodesInTree,LeafNodes);
	      printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	      fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	      }
	     break;

  case 'n' : writeStackNode(TTTfile,lastTrans->mainroot);
	     fprintf(stdout, "\nNew transaction .");
	     fprintf(stdout, " new transaction-no : %d .\n" , CountTranspos+1);
	     fprintf(stdout, "");
	     fprintf(PRINTfile, "\nNew transaction .");
	     fprintf(PRINTfile, " new transaction-no : %d .\n" , CountTranspos+1);
	     fprintf(PRINTfile, "");
	     getmemForSTACK(&firstTrans,&lastTrans);
	     break;

  case 'q' : flag=TRUE;

  }
 }
while ( !flag );

writeStackNode(TTTfile,lastTrans->mainroot);
freeSTACKlist(lastTrans);
firstTrans = lastTrans =NULL;
printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
fcloseall();

}  /* void main */

/*------------------------------------------------------------------------*/

void initCounters(int *CountforTrans,int *CountforRecs)
{
/********************************************\
| output: the file counters                  |
\********************************************/

 int  handle1, handle2;

 handle1 = open("trans.dat",O_RDONLY,O_BINARY);
 *CountforTrans = filelength(handle1)/(sizeof(STACK));
 close(handle1);

 handle2 = open("rec.dat",O_RDONLY,O_BINARY);
 *CountforRecs = filelength(handle2)/(sizeof(BTREE));
 close(handle2);
}

/*------------------------------------------------------------------------*/

void initSTACKlist(STACKLINK **f, STACKLINK **m, FILE *TTTfile)
{
/************************************************\
| input:  the TTT file     	                 |
| output: the double linked list of T-tree roots |
\************************************************/

STACKLINK *new;
STACK     a;
int i;

for(i=1;i<=CountTranspos;i++)
 {
 if( (new=(STACKLINK *)malloc(sizeof(STACKLINK)))==NULL )
  {
  fprintf(stdout,"Memory allocation failed\n");
  exit(1);
  }
  new->mainroot = getStackNode(TTTfile,i);
  if( (*m)!=NULL )
   {
   new->prev = *m;
   new->next = NULL;
   (*m)->next = new;
   *m = new;
   }
  else
   {
   new->prev = NULL;
   new->next = NULL;
   *f = new;
   *m = new;
   }
  }

}

/*------------------------------------------------------------------------*/

void batch_modify(SL **first, int *rpos, FILE *RECfile)
{
/*************************************************************\
| This is the main routine of the modification process in the |
| Overlapping B+-trees                                        |
|                                                             |
| input:  the root of the T-tree in process                   |
|	  and the linked list of records                      |
|	  for batch insertion, deletion or update 	      |
| output: the root of the T-tree in process     	      |
\*************************************************************/

REC      k;
BTREE    p, buffer[ARRAYsize];
int      eofkeys = FALSE,
	 mergeflag =FALSE;
int      level, h, i;
long     decisionvalue;
long     lastkeymodifinW;


for(i=0;i<=ARRAYsize-1;i++)
 {
 w[i].recontain=w[i].contain=0;   /* global w */
 inter(w[i]).savedkey=0;
// w[i].data=FALSE;
 buffer[i].data=FALSE;
 buffer[i].contain=0;
  }
level=h=1;

if( (*rpos)!=NULLPOS )
 {
 p = getBTreeNode(RECfile, (long)(*rpos));
 buffer[base+h]=p;
  if( isLEAF(p) )
   {
   exter(w[base+h]).rentpos=++CountRecpos;
   exter(w[base+h]).savedbpoint=NULLPOS;
   setpointers(&(buffer[base+h]), &(w[base+h]));
   }
 insertBTREEnodeinW(&(w[base+h]), p);

 k=getnextkey(first,&eofkeys);

 while(!eofkeys)
  {
  if(!isLEAF(w[base+level]))
   {
   mergeflag =FALSE;
   h = level;
   BTree_path(w, k.key, &h, buffer, RECfile);
   }
  level = h;

  if(!eofkeys)
   decisionvalue=kfieldinw(buffer,level);

  while( (!eofkeys) && (k.key<=decisionvalue) )
   {
   printf("%ld   %c.....",k.key,k.tag);
   modifykinWh(&(w[base+h]),k);
   lastkeymodifinW=k.key;
   if(w[base+h].contain==WMAXex)
     split(h, w, buffer, RECfile, FALSE);

   k=getnextkey(first,&eofkeys);
   }      /* (!eofkeys) && (k.g.key<decisionvalue) */

  if(!eofkeys)
   while( (k.key>decisionvalue) )
    {
    freeW(&level, w, buffer, lastkeymodifinW, rpos, &mergeflag, RECfile);
    decisionvalue=kfieldinw(buffer, level);
    }
  }   //while(!eofkeys)

 while(base+level>0)
  freeW(&level, w, buffer, lastkeymodifinW, rpos, &mergeflag, RECfile);
 }  /*if( mainroot->treepointer!=NULLPOS ) */
else
 {
 w[base+h].data=TRUE;
 exter(w[base+h]).rentpos=++CountRecpos;
 exter(w[base+h]).savedbpoint=NULLPOS;

 k=getnextkey(first,&eofkeys);
 while(!eofkeys)
  {
  decisionvalue=MAXLONG;
  while(!eofkeys)
   {
   printf("%ld   %c.....",k.key,k.tag);
   modifykinWh(&(w[base+h]),k);
   lastkeymodifinW=k.key;
   if(w[base+h].contain==WMAXex)
     split(h, w, buffer, RECfile, FALSE);
   k=getnextkey(first,&eofkeys);
   }
  } 	 //while(!eofkeys)
 while(base+level>0)
  freeW(&level, w, buffer, lastkeymodifinW, rpos, &mergeflag, RECfile);
 }  /*if( mainroot->treepointer==NULLPOS ) */
}

/*------------------------------------------------------------------------*/

void  setpointers(BTREE *bp, Workspace *w)
{
/*************************************************************\
| Connect a historical leaf node '*bp' with its new copy node |
| (the Fpoint of the historical node points to the new node   |
| and the Bpoint of the new node points to the old node)      |
|                                                             |
| input:  the historical node *bp            	              |
| output: the new node which is now in the wordspace node *w  |
\*************************************************************/

 if( bp->starttime<CountTranspos )
   {
   external(*bp).Fpoint=exter(*w).rentpos;;
   exter(*w).savedbpoint=bp->pos;
   }

}

/*------------------------------------------------------------------------*/

void split(int lev,Workspace w[], BTREE buffer[], FILE *RECfile, int fromfree)
{
/*************************************************************\
| Splits a workspace node and writes it in the disk           |
|                                                             |
| input:  the workspace node, its historical buffer[] node    |
|         and a flag determining if the call                  |
|	  of the routine split was made by the routine 'free' |
| output: the workspace node                                  |
\*************************************************************/

item  k;
BTREE p;

if( !isLEAF(w[base+lev]) )
 restoreBTREEnodefromWi(&p, &(w[base+lev]), MAXin, &k);
else
 {
 restoreBTREEnodefromWi( &p, &(w[base+lev]), MAXex, &k);

 if( NOTFIRSTTRANS )
  {
  external(p).Bpoint = exter(w[base+lev]).savedbpoint;
  external(p).FCpoint =exter(w[base+lev]).rentpos;
  }
 }

if( isLEAF(w[base+lev]) && (!fromfree) )
 emptybuffer(buffer[base+lev], RECfile);

writeBTreeNode(RECfile, p);

insertRECinW(&(w[base+lev-1]), k);
if(w[base+lev-1].contain==WMAXin)
 split(lev-1, w, buffer, RECfile, fromfree);

}

/*------------------------------------------------------------------------*/

void restoreBTREEnodefromWi(BTREE *p, Workspace *w, int numKeys, item *k)
{
/***********************************************************\
| Restores a new Overlapping B+-tree node with numKeys keys |
| from the workspace node *w.                               |
| k is the last (maximum) key(heading key)                  |
| of the node with a pointer to its node position     	    |
\***********************************************************/

 int i;

 newBTreenode(p, w, numKeys, k);

 if( isLEAF(*w) )
  {
  for(i=numKeys+1; i<=w->contain; i++)
   {
   exter(*w).key[i-numKeys] = exter(*w).key[i];
   exter(*w).seckey[i-numKeys] = exter(*w).seckey[i];
   }
  w->contain = w->contain-numKeys;

  exter(*w).rentpos = ( w->contain==0 ? NULLPOS : ++CountRecpos );

  if( (w->contain==0) && (external(*p).key[numKeys]<inter(*(w-1)).savedkey) )
   maxkeyfor(*p) = inter(*(w-1)).savedkey;
  else
   maxkeyfor(*p) = external(*p).key[numKeys];

  }
 else
  {
  for(i=numKeys+1; i<=w->contain; i++)
   {
   inter(*w).key[i-numKeys-1] = inter(*w).key[i];
   inter(*w).p[i-numKeys-1] = inter(*w).p[i];
   }
  w->recontain=w->recontain-numKeys-1;
  w->contain=(w->recontain==0 ? w->contain-numKeys : w->contain-numKeys-1);
  }

}

/*------------------------------------------------------------------------*/

void newBTreenode(BTREE *r, Workspace *w, int numKeys, item *k)
{
/*****************************************************************\
| Builds a new Overlapping B+tree node from the workspace node *w |
|                                                                 |
| output: the Overlapping B+tree node and the item *k which       |
|         represents the pair <maximum key, disk address>         |
|         of Overlapping B+tree node			          |
\*****************************************************************/

int i;

 r->data    = w->data;
 r->starttime = CountTranspos;
 r->contain = 0;

 if( isLEAF(*r) )
  {
  r->pos     = exter(*w).rentpos;
  external(*r).endtime  = MAXINT;
  }
 else
  r->pos     = ++CountRecpos;

 if( isLEAF(*r) )
  {
  for (i=1;i<=numKeys;i++)
   {
   external(*r).key[i]=exter(*w).key[i];
   external(*r).seckey[i]=exter(*w).seckey[i];
   r->contain++;
   }
  external(*r).Fpoint=NULLPOS;
  external(*r).FCpoint=NULLPOS;
  external(*r).Bpoint=NULLPOS;
  external(*r).BCpoint=NULLPOS;

  k->key=external(*r).key[r->contain];
  k->p=r->pos;
  }
 else
  {
  for (i=0;i<=numKeys;i++)
   {
   internal(*r).key[i]=inter(*w).key[i];
   internal(*r).p[i]=inter(*w).p[i];
   }
  r->contain = numKeys;

  k->key=internal(*r).key[r->contain];  /*  (*r).key[0] 夜 ⤘ 姜  w */
  k->p=r->pos;
  }
}

/*------------------------------------------------------------------------*/

void insertBTREEnodeinW(Workspace *w,BTREE q)
{
/**********************************************************\
| Copy an Overlapping B+tree node in the workspace node *w |
\**********************************************************/

int i;
int pos,found;

 w->data=q.data;

 if ( !isLEAF(q) )
  {
  bin_searchinW(*w,internal(q).key[0],&pos,&found);
  if( (w->recontain!=0) && (inter(*w).key[0]<internal(q).key[0]) )
    pos++;

  for(i=w->recontain+q.contain;i>=pos+q.contain+1;i--)
   {
   inter(*w).key[i]=inter(*w).key[i-q.contain-1];
   inter(*w).p[i]=inter(*w).p[i-q.contain-1];
   }
  for(i=pos;i<=(q.contain+pos);i++)
   {
   inter(*w).key[i]=internal(q).key[i-pos];
   inter(*w).p[i]=internal(q).p[i-pos];
   }
  w->contain=(w->recontain==0 ? w->contain+q.contain : w->contain+q.contain+1);
  w->recontain=w->recontain + q.contain + 1;
  }
 else   /* isLEAF(q) */
  {
  bin_searchinW(*w,external(q).key[1],&pos,&found);

  for(i=w->contain+q.contain;i>=pos+q.contain;i--)
   {
    exter(*w).key[i]=exter(*w).key[i-q.contain];
    exter(*w).seckey[i]=exter(*w).seckey[i-q.contain];
   }
  for(i=pos;i<=(q.contain+pos-1);i++)
   {
   exter(*w).key[i]=external(q).key[i-pos+1];
   exter(*w).seckey[i]=external(q).seckey[i-pos+1];
   }
  w->contain=w->contain + q.contain;
  }
}

/*------------------------------------------------------------------------*/

void emptybuffer(BTREE buffer, FILE *RECfile)
{
/**************************************\
| The node 'buffer' becomes historical |
| by setting its endtime field         |
\**************************************/

  if( (NOTFIRSTTRANS) && (buffer.contain!=0) )
   {
   external(buffer).endtime=CountTranspos-1;
   writeBTreeNode(RECfile, buffer);
   }
}

/*------------------------------------------------------------------------*/

void  freeW(int *lev, Workspace w[], BTREE buffer[], long lastk,
	    int *rpos, int *LEAFmerge, FILE *RECfile)
{
/*************************************************************\
| It empties the workspace node of the *lev level of the tree |
| and it wirtes this node in the disk                         |
| If this level represents the root, then it updates the      |
| root pointer *rpos, also.                                   |
\*************************************************************/

item       k1,k2;
BTREE      p1,p2,q;
int        i,qpos;
long int   tempkey;

if(!isLEAF(w[base+(*lev)]))
 {
 if(w[base+(*lev)].contain>2*MAXin)
  split(*lev, w, buffer, RECfile, TRUE);

 if(w[base+(*lev)].contain>MAXin)
  {
   restoreBTREEnodefromWi(&p1, &(w[base+(*lev)]), w[base+(*lev)].contain-( w[base+(*lev)].contain + 1 )/2, &k1);
   inter(w[base+(*lev)-1]).savedkey = 0;
   restoreBTREEnodefromWi(&p2, &(w[base+(*lev)]), w[base+(*lev)].contain, &k2);

  writeBTreeNode(RECfile, p1);
  writeBTreeNode(RECfile, p2);

  tempkey = inter(w[base+(*lev)-1]).savedkey;
  inter(w[base+(*lev)-1]).savedkey = 0;
  insertRECinW(&(w[base+(*lev)-1]), k1);
  if(w[base+(*lev)-1].contain==WMAXin)
   split((*lev)-1, w, buffer, RECfile, TRUE);

  inter(w[base+(*lev)-1]).savedkey = tempkey;
  insertRECinW(&(w[base+(*lev)-1]), k2);
  if(w[base+(*lev)-1].contain==WMAXin)
   split((*lev)-1, w, buffer, RECfile, TRUE);
  (*lev)--;
  }  /*  if(w[base+(*lev)].contain>MAXin)  */
 else if(w[base+(*lev)].contain>=MINin)
  {
  if( isDifferent(w[base+(*lev)],buffer[base+(*lev)]) )
   {
   restoreBTREEnodefromWi(&p1, &(w[base+(*lev)]), w[base+(*lev)].contain, &k1);
   writeBTreeNode(RECfile, p1);
   *rpos = p1.pos;
   }
  else
   {
   k1.key = internal(buffer[base+(*lev)]).key[0];
   k1.p   = buffer[base+(*lev)].pos;
   w[base+(*lev)].recontain=w[base+(*lev)].contain=0;
   }
  insertRECinW(&(w[base+(*lev)-1]), k1);
  if(w[base+(*lev)-1].contain==WMAXin)
   split((*lev)-1, w, buffer, RECfile, TRUE);
  (*lev)--;
  }
 else /* if(w[base+(*lev)].contain<MINin) */
  {
  i=base+(*lev)-1;
  while( (w[i].contain==0) && (w[i].recontain==0) && (i>0) )
   i--;
  if( (w[i].recontain==0) )
   {
   if(w[base+(*lev)].recontain!=0) // we don't need the if
    {
     if(w[base+(*lev)].contain!=0)
      {
      restoreBTREEnodefromWi(&p1, &(w[base+(*lev)]), w[base+(*lev)].contain, &k1);
      writeBTreeNode(RECfile, p1);
      *rpos = p1.pos;
      }
     else
      *rpos=inter(w[base+(*lev)]).p[0];
     *lev=-base;
    } // if(w[base+(*lev)].recontain!=0)
   }
  else // w[i].recontain!=0
   {
   while( i<base+(*lev) )
    {
    q = restore(&(w[i]), &qpos, lastk, &(buffer[i+1]), RECfile);
    k1.key=inter(w[i]).key[qpos];
    k1.p=inter(w[i]).p[qpos];

    insertBTREEnodeinW(&(w[++i]), q);
    removeRECinW(&(w[i-1]),k1);
    }
   }  /* w[i].recontain!=0              */
  }   /* if(w[base+(*lev)].contain<MINin) */
 }    /* if(!isLEAF(w[base+(*lev)])) */

else  /* if( isLEAF(w[base+(*lev)])) */
 {
 if(w[base+(*lev)].contain>2*MAXex)
  split(*lev, w, buffer, RECfile, TRUE);

 if(w[base+(*lev)].contain>MAXex)
  {
  restoreBTREEnodefromWi(&p1, &(w[base+(*lev)]), w[base+(*lev)].contain-( w[base+(*lev)].contain + 1 )/2, &k1);
  restoreBTREEnodefromWi(&p2, &(w[base+(*lev)]), w[base+(*lev)].contain, &k2);

  if( NOTFIRSTTRANS )
   {
   external(p1).Bpoint = exter(w[base+(*lev)]).savedbpoint;
   external(p2).Bpoint = exter(w[base+(*lev)]).savedbpoint;
   external(p1).FCpoint = p2.pos;
   }
  writeBTreeNode(RECfile, p1);
  writeBTreeNode(RECfile, p2);
  emptybuffer(buffer[base+(*lev)], RECfile);

  tempkey = inter(w[base+(*lev)-1]).savedkey;
  inter(w[base+(*lev)-1]).savedkey = 0;
  insertRECinW(&(w[base+(*lev)-1]), k1);
  if(w[base+(*lev)-1].contain==WMAXin)
   split((*lev)-1, w, buffer, RECfile, TRUE);

  inter(w[base+(*lev)-1]).savedkey = tempkey;
  insertRECinW(&(w[base+(*lev)-1]), k2);
  if(w[base+(*lev)-1].contain==WMAXin)
   split((*lev)-1, w, buffer, RECfile, TRUE);
  (*lev)--;
  }  /*  if(w[base+(*lev)].contain>MAXex)  */
 else if(w[base+(*lev)].contain>=MINex)
  {
  if( isDifferent(w[base+(*lev)],buffer[base+(*lev)]) || (*LEAFmerge) )
   {
   restoreBTREEnodefromWi(&p1, &(w[base+(*lev)]), w[base+(*lev)].contain, &k1);

   external(p1).Bpoint  = exter(w[base+(*lev)]).savedbpoint;
   exter(w[base+(*lev)]).savedbpoint = NULLPOS;

   writeBTreeNode(RECfile, p1);
   emptybuffer(buffer[base+(*lev)], RECfile);

   *rpos = p1.pos;
   }
  else
   {
   exter(w[base+(*lev)]).savedbpoint = NULLPOS;
   if( exter(w[base+(*lev)]).rentpos!=NULLPOS)
    {
    CountRecpos--;
    exter(w[base+(*lev)]).rentpos=NULLPOS;
    }
   k1.key = external(buffer[base+(*lev)]).key[1];
   k1.p   = buffer[base+(*lev)].pos;
   w[base+(*lev)].contain=0;
   }
  insertRECinW(&(w[base+(*lev)-1]), k1);
  if(w[base+(*lev)-1].contain==WMAXin)
   split((*lev)-1, w, buffer, RECfile, TRUE);
  (*lev)--;
  }
 else /* if(w[base+(*lev)].contain<MINex) */
  {
  i=base+(*lev)-1;
  while( (w[i].contain==0) && (w[i].recontain==0) && (i>0) )
   i--;
  if( w[i].recontain==0 )
   {
    if(w[base+(*lev)].contain!=0)
     {
      restoreBTREEnodefromWi(&p1, &(w[base+(*lev)]), w[base+(*lev)].contain, &k1);

      external(p1).Bpoint  = exter(w[base+(*lev)]).savedbpoint;
      exter(w[base+(*lev)]).savedbpoint = NULLPOS;

      writeBTreeNode(RECfile, p1);
      emptybuffer(buffer[base+(*lev)], RECfile);
      *rpos = p1.pos;
     }
    else
      *rpos = NULLPOS;
    *lev=-base;
   }
  else // w[i].recontain!=0
   {
   while( i<base+(*lev) )
    {
    q = restore(&(w[i]), &qpos, lastk, &(buffer[i+1]), RECfile);
    *LEAFmerge = TRUE;
    k1.key=inter(w[i]).key[qpos];
    k1.p=inter(w[i]).p[qpos];

    insertBTREEnodeinW(&(w[++i]), q);
    removeRECinW(&(w[i-1]),k1);
    }
   }  /* w[i].recontain!=0                */
  }   /* if( w[base+(*lev)].contain<MINex ) */
 }    /* if( isLEAF(w[base+(*lev)]) )       */
}

/*------------------------------------------------------------------------*/

BTREE restore(Workspace *w, int *qpos, long lastkinW, BTREE *buffer, FILE *RECfile)
{
/*************************************************************\
| It searchs for a sibling T-tree node of the node            |
| which is in the *buffer.      	                      |
| Ihis routine is called when the workspace node underflows.  |
\*************************************************************/

BTREE q, ftemp,
      btemp;
int   found;
int   fromright;

bin_searchinW(*w,lastkinW,qpos,&found);

q = getBTreeNode(RECfile, (long)(inter(*w).p[*qpos]));

if( (w->contain==0) || (*qpos)==w->contain )
 fromright=( inter(*w).key[*qpos] < lastkinW ? FALSE : TRUE );
else // if( (w->contain!=0) || (*qpos)!=w->contain )
 fromright=TRUE;


w++;

if( NOTFIRSTTRANS )
 {
 if(q.starttime<CountTranspos)
  {
  if(isLEAF(q))
   {
   if(fromright)
    {
     external(*buffer).BCpoint=q.pos;
     emptybuffer(*buffer, RECfile);
     external(q).Fpoint=exter(*w).rentpos;
     *buffer = q;
    }
   else //if(!fromright)
    {
     external(q).BCpoint=exter(*w).savedbpoint; // I think ...point=exter(*w).savedbpoint;
     external(q).Fpoint=exter(*w).rentpos;
     emptybuffer(q, RECfile);
     exter(*w).savedbpoint=q.pos;
    }
   }
  else //if(!isLEAF)
   {
   if(fromright)
    *buffer = q;
   }
  }
 else // if(q.starttime==CountTranspos)
  {
  if(isLEAF(q))
   { 						  //  夜  夘 fromright  ᚦ ⮜   
   btemp = getBTreeNode(RECfile, (long)external(q).Bpoint);
   if(external(q).Bpoint!=exter(*w).savedbpoint) //1.
    {
    if(external(btemp).Fpoint!=q.pos) //1b.
     {
     ftemp = getBTreeNode(RECfile, (long)external(btemp).Fpoint);
     while(external(ftemp).FCpoint!=q.pos)
       ftemp = getBTreeNode(RECfile, (long)external(ftemp).FCpoint);
     external(ftemp).FCpoint = exter(*w).rentpos;
     writeBTreeNode(RECfile, ftemp);
     while( (external(btemp).BCpoint!=NULLPOS) && (external(btemp).Fpoint!=q.pos) )
      btemp = getBTreeNode(RECfile, (long)external(btemp).BCpoint);
     }

    while(external(btemp).Fpoint==q.pos) // 1a  1b
     {
     external(btemp).Fpoint = exter(*w).rentpos;
     writeBTreeNode(RECfile, btemp);
     if(external(btemp).BCpoint==NULLPOS)
      break;
     btemp = getBTreeNode(RECfile, (long)external(btemp).BCpoint);
     }
    if(external(btemp).BCpoint==NULLPOS)
     {
     external(btemp).BCpoint = exter(*w).savedbpoint;
     writeBTreeNode(RECfile, btemp);
     }
    exter(*w).savedbpoint = external(q).Bpoint;
    }

   else // if(external(q).Bpoint==exter(*w).savedbpoint)  //2
    {
    if(external(btemp).Fpoint!=q.pos)  //2b.
     {
     ftemp = getBTreeNode(RECfile, (long)external(btemp).Fpoint);
     while(external(ftemp).FCpoint!=q.pos)
       ftemp = getBTreeNode(RECfile, (long)external(ftemp).FCpoint);
     external(ftemp).FCpoint = exter(*w).rentpos;
     writeBTreeNode(RECfile, ftemp);
     while( (external(btemp).BCpoint!=NULLPOS) && (external(btemp).Fpoint!=q.pos) )
      btemp = getBTreeNode(RECfile, (long)external(btemp).BCpoint);
     }

    while(external(btemp).Fpoint==q.pos)  // 2a  2b
     {
     external(btemp).Fpoint = exter(*w).rentpos;
     writeBTreeNode(RECfile, btemp);
     if(external(btemp).BCpoint==NULLPOS)
      break;
     btemp = getBTreeNode(RECfile, (long)external(btemp).BCpoint);
     }
    }
   }  // if(isLEAF(q))
  }   // if(q.starttime==CountTranspos)
 }    // if( NOTFIRSTTRANS )
return(q);
}

/*------------------------------------------------------------------------*/

int isDifferent(Workspace w,BTREE b)
{
/*************************************************\
| It compares two T-tree nodes of the same level. |
| One of the nodes is the workspace node w.        |
\*************************************************/

int i;

if(!b.contain)
 return(TRUE);
if(w.contain!=b.contain)
 return(TRUE);

if(!isLEAF(w))
 {
 for(i=0;i<=w.contain;i++)
  if( (inter(w).key[i]!=internal(b).key[i]) ||
      (inter(w).p[i]!=internal(b).p[i]) )
	 return(TRUE);
 }
else /* if(isLEAF(w)) */
 {
 for(i=1;i<=w.contain;i++)
  if( (exter(w).key[i]!=external(b).key[i]) ||
      (exter(w).seckey[i]!=external(b).seckey[i]) )
	return(TRUE);
 }

return(FALSE);
}

/*------------------------------------------------------------------------*/

long kfieldinw(BTREE buffer[],int height)
{
/********************************************\
| It searches for the maximum key which      |
| is or could be stored in a T-tree node     |
\********************************************/

int  i, found, pos;
long decisionkey, x;

decisionkey=MAXLONG;
if( !isLEAF(buffer[base+height]) )
 x = internal(buffer[base+height]).key[buffer[base+height].contain];
else
 x = external(buffer[base+height]).key[buffer[base+height].contain];

for(i=base+height-1;i>=0;i--) /* only internal w */
 {
 bin_search(buffer[i],x,&pos,&found);
 if( pos!=(buffer[i].contain) )
  {
  decisionkey=internal(buffer[i]).key[pos];
  break;
  }
 }
return(decisionkey);
}

/*------------------------------------------------------------------------*/

void modifykinWh(Workspace *w,REC k)
{
item v;

 v.key  = k.key;
 v.seckey = rand();

 switch(k.tag)
  {
  case 'i' : insertRECinW(w, v);
	     break;

  case 'd' : removeRECinW(w,v);
	     break;

  case 'u' : updateRECinW(w,v);
	     break;
  }

}

/*------------------------------------------------------------------------*/

void insertRECinW(Workspace *w, item g)
{
/**************************************************\
| It inserts a new record in the workspace node *w |
\**************************************************/

 int   i;
 int   found,
       pos;

 bin_searchinW(*w,g.key,&pos,&found);
 if( !isLEAF(*w) )
 {
 if( !found )
  {
  if( (w->recontain!=0) && (w->contain==pos) )
   pos = ( inter(*w).key[w->contain]<g.key ? pos+1 : pos );
  for(i=w->recontain-1;i>=pos;i--)
   {
   inter(*w).key[i+1] = inter(*w).key[i];
   inter(*w).p[i+1]   = inter(*w).p[i];
   }

  if( (w+1)->contain==0 )
   {
   if( g.key>=inter(*w).savedkey )
    inter(*w).key[pos] = g.key;
   else
    inter(*w).key[pos] =inter(*w).savedkey;
   }
  else
   inter(*w).key[pos] = g.key;

  inter(*w).p[pos]   = g.p;
  inter(*w).savedkey = 0;

  w->contain=(w->recontain==0 ? w->contain : w->contain+1);
  w->recontain++;
  }
 else /* if(found) */
  inter(*w).p[pos]=g.p;
 }
else /* isLEAF */
 {
 if( !found ) 
  {
  for(i=w->contain;i>=pos;i--)
   {
    exter(*w).key[i+1]=exter(*w).key[i];
    exter(*w).seckey[i+1]=exter(*w).seckey[i];
   }
   exter(*w).key[pos]=g.key;
   exter(*w).seckey[pos]=g.seckey;
  (w->contain)++;
  fprintf(stdout,"Inserted key %ld .\n",g.key);
  }
 else  /* if(found) */
  fprintf(stdout,"Inserted key %ld allready exists.\n",g.key);
 }
}

/*------------------------------------------------------------------------*/

void removeRECinW(Workspace *w, item g)
{
/************************************************\
| It deletes a record from the workspace node *w |
\************************************************/

 int   i;
 int   found,
       pos;

 bin_searchinW(*w,g.key,&pos,&found);
 if( found )
  {
  if( !isLEAF(*w) )
   {
   if( g.key > inter(*w).savedkey )
    inter(*w).savedkey = g.key;

   for(i=pos+1; i<=w->contain; i++)
    {
    inter(*w).key[i-1] = inter(*w).key[i]; /*  ⧨    p.[pos-1] (I think no) */
    inter(*w).p[i-1] = inter(*w).p[i];
    }
   w->contain=(w->recontain==1 ? w->contain : w->contain-1);
   w->recontain--;
   }
  else /* isLEAF */
   {
   for(i=pos+1; i<=w->contain; i++)
    {
    exter(*w).key[i-1] = exter(*w).key[i];
    exter(*w).seckey[i-1] = exter(*w).seckey[i];
    }
   fprintf(stdout,"Deleted  key %ld .\n",g.key);
   (w->contain)--;
   }
  } /* found */
 else if(isLEAF(*w))  /* and !found */
  fprintf(stdout,"Deleted  key %ld does not exists.\n",g.key);

}  // removeRec()

/*------------------------------------------------------------------------*/

void updateRECinW(Workspace *w, item g)
{
/**********************************************\
| It updates a record of the workspace node *w |
\**********************************************/

int   found,
      pos;

bin_searchinW(*w,g.key,&pos,&found);

if( found )
 {
 if(!isLEAF(*w))
  inter(*w).p[pos]=g.p;
 else /* isLEAF(*w) */
  {
  exter(*w).seckey[pos]=rand();
  fprintf(stdout,"Updated  key %ld .\n",g.key);
  }
 }
else if(isLEAF(*w))  /* and !found */
  fprintf(stdout,"Updated  key %ld does not exists.\n",g.key);
}

/*------------------------------------------------------------------------*/

void BTree_path(Workspace w[], long x, int *height, BTREE buffer[], FILE *RECfile)
{
/********************************************************************************\
| It inserts a whole T-tree path (root to specific leaf) in the workspace array. |
| Driver for the accessing of the correct path is the key x                      |
\********************************************************************************/

BTREE p;
int   pathpos,
      found;
item  g;

bin_searchinW(w[base+(*height)], x, &pathpos, &found);

g.key = inter(w[base+(*height)]).key[pathpos];
g.p   = inter(w[base+(*height)]).p[pathpos];
removeRECinW(&(w[base+(*height)]),g);
p = getBTreeNode(RECfile, (long)(g.p));

(*height)++;
insertBTREEnodeinW(&(w[base+(*height)]), p);
buffer[base+(*height)]=p;

if(!isLEAF(p))
 BTree_path(w, x, height, buffer, RECfile);
else
 {
 exter(w[base+(*height)]).rentpos=++CountRecpos;
 exter(w[base+(*height)]).savedbpoint=NULLPOS;
 setpointers(&(buffer[base+(*height)]), &(w[base+(*height)]));
 }
}

/*------------------------------------------------------------------------*/

void bin_searchinW(Workspace  w,long newrec,int *pos,int *found)
{
/**********************************************\
| A fast binary search in the workspace node w |
\**********************************************/

int l,k;

l=1;
*pos=w.contain;

if( !isLEAF(w) )
 {
 if(w.recontain>1)
  {
  if( newrec<=inter(w).key[0] )
   {
    *pos   = 0;
    *found = (newrec==inter(w).key[*pos]);
    return ;
   }
  do
   {
   k=(l+*pos)/2;
   if (newrec<=inter(w).key[k])  *pos=k-1;
   if (newrec>=inter(w).key[k])   l=k+1;
   }
  while (*pos>=l);
  *found=(l-*pos>1);
  if(*pos!=w.contain)
   (*pos)++;
  }
 else
  {
  *pos = 0;
  *found=(w.recontain==1 ? (newrec==inter(w).key[0]) : FALSE);
  return;
  }
 }
else
 {
 if(w.contain!=0)
  {
  do
   {
   k=(l+*pos)/2;
   if (newrec<=exter(w).key[k])  *pos=k-1;
   if (newrec>=exter(w).key[k])   l=k+1;
   }
  while (*pos>=l);
  *found=(l-*pos>1);
  (*pos)++;
  }
 else
  {
   *pos   = 1;
   *found = FALSE;
    return ;
  }
 }
}

/*------------------------------------------------------------------------*/

void bin_search(BTREE a,long newrec,int *pos,int *found)   /**found:boolean*/
{
/*******************************************\
| A fast binary search in the T-tree node a |
\*******************************************/

int l,k;

l=1;
*pos=a.contain;
if( !isLEAF(a) )
 {
  if( newrec<=internal(a).key[0] )
   {
    *pos   = 0;
    *found = (newrec==internal(a).key[*pos]);
    return ;
   }
 do
  {
  k=(l+*pos)/2;
  if (newrec<=internal(a).key[k])  *pos=k-1;
  if (newrec>=internal(a).key[k])   l=k+1;
  }
 while (*pos>=l);
 *found=(l-*pos>1);
 if(*pos!=a.contain)
  (*pos)++;
 }
else
 {
 do                   
  {
  k=(l+*pos)/2;
  if (newrec<=external(a).key[k])  *pos=k-1;
  if (newrec>=external(a).key[k])   l=k+1;
  }
 while (*pos>=l);
 *found=(l-*pos>1);
 (*pos)++;
 }
}

/*------------------------------------------------------------------------*/

void search(STACKLINK *first, STACKLINK *last, FILE *RECfile, FILE *PRINTfile)
{
/***********************************************************************\
| This is the menu for the different queries.                           |
| The user is called to specify:                                        |
| 1. the type of query                                                  |
| 2. the way that the algorithm will search                             |
|    (starting from the first time point and moving forwards            |
|    or starting from the last time point and moving in a backward way)	|
|                                               			|
| Note: 'first' and 'last' are the two pointers 			|
|       of the double linked list of the roots  			|
|	of the different T-trees     	        			|
\***********************************************************************/

 char        query;

 fprintf(stdout, "\n1.Find the values of all the keys during the T time point.");
 fprintf(stdout, "\n2.Find the values of all the keys between Xi->Xj keys during the Ti time point");
 fprintf(stdout, "\n3.Find the history of the key X in the range between Ti->Tj time points.");
 fprintf(stdout, "\n4.Find the history of the key X in the oposite range between Ti<-Tj time points");
 fprintf(stdout, "\n5.Find the history of Y keys (range) starting from the key X between Ti->Tj time points.");
 fprintf(stdout, "\n6.Find the history of Y keys (range) starting from the key X between Ti<-Tj time points.");
 fprintf(stdout, "\n7.Find all the history of the key X going forwards.");
 fprintf(stdout, "\n8.Find all the history of the key X going backwards.");
 fprintf(stdout, "\n9.Find all the history of Y keys (range) starting from the key X going forwards");
 fprintf(stdout, "\n0.Find all the history of Y keys (range) starting from the key X going backwards");
 fprintf(stdout, "\n-------------------------------------------------------------------------------");
 fprintf(stdout, "\nSelect query : ");
 query = getch();
 fprintf(stdout, "%c",query);
 fprintf(PRINTfile, "\nSelect query : %c",query);

 switch( query )
  {

  case '1':
  case '2' : TimesliceQuery(query, first, RECfile, PRINTfile);
	     break;

  case '3' :
  case '4' :
  case '5' :
  case '6' : PurekeyQueryWithPredictTime(query, first, last, RECfile, PRINTfile);
	     break;

  case '7' :
  case '8' :
  case '9' :
  case '0' : PurekeyQuery(query, first, last, RECfile, PRINTfile);
	     break;
  } // switch

}

/*------------------------------------------------------------------------*/

void PurekeyQueryWithPredictTime(char ch, STACKLINK *first, STACKLINK *last, FILE *RECfile, FILE *PRINTfile)
{
/**************************************************\
| The user is called to specify the time interval  |
| [trans1, strans2] in which he/she is interesting |
| to search, for the history of a specific key     |
| (pure key query)				   |
\**************************************************/

 int    trans1,trans2;

 do
  {
  fprintf(stdout,"\nGive me the range of transactions to search [0 0 = end] : ");
  fscanf(stdin, "%d %d",&trans1,&trans2);
  fprintf(stdout, "\n");
  fprintf(PRINTfile,"\nGive me the range of transactions to search [0 0 = end] : %d %d\n",trans1,trans2);
  if( (trans1!=0) || (trans2!=0) )
   {
   if( (first->mainroot.transaction>trans2) || (last->mainroot.transaction<trans1))
    {
    fprintf(stdout, "Wrong group of transactions .\n");
    fprintf(PRINTfile, "Wrong group of transactions .\n");
    }
   else
    KindOfPurekeyQuery(ch, first, last, trans1, trans2, RECfile, PRINTfile);
   }
  }
 while ( (trans1!=0) || (trans2!=0) );


}

/*------------------------------------------------------------------------*/

void PurekeyQuery(char ch, STACKLINK *first, STACKLINK *last, FILE *RECfile, FILE *PRINTfile)
{

 int    trans1,trans2;

 trans1 = first->mainroot.transaction;
 trans2 = last->mainroot.transaction;
 ch = (ch=='7' ? '3' : ( ch=='8' ? '4' : ( ch=='9' ? '5' : '6' ))) ;

 fprintf(stdout, "\n");
 fprintf(PRINTfile, "\n");
 KindOfPurekeyQuery(ch, first, last, trans1, trans2, RECfile, PRINTfile);

}

/*------------------------------------------------------------------------*/

void KindOfPurekeyQuery(char ch, STACKLINK *first, STACKLINK *last,
			int trans1, int trans2, FILE *RECfile, FILE *PRINTfile)
{
/***********************************************************************\
| The user is called to specify the key 'x' or the range of keys        |
| whose history he/she is interesting to search for                     |
\***********************************************************************/

long   x;
int    numberkeys;

switch( ch )
 {
 case '3': do
	    {
	    fprintf(stdout, "Give me the key whose history you ask for [0=end] : ");
	    fscanf(stdin, "%ld",&x);
	    fprintf(stdout, "\n");
	    fprintf(PRINTfile, "Give me the key whose history you ask for [0=end] : %ld\n",x);
	    if( x!=0 )
	      {
	      historywithFpoints(first, last, x, trans1, trans2, 1, RECfile, PRINTfile);
	      printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	      fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	      }
	    fprintf(stdout, "\n");
	    fprintf(PRINTfile, "\n");
	    }
	   while ( (x!=0) );
	   break;

 case '4': do
	    {
	    fprintf(stdout, "Give me the key whose history you ask for [0=end] : ");
	    fscanf(stdin, "%ld",&x);
	    fprintf(stdout, "\n");
	    fprintf(PRINTfile, "Give me the key whose history you ask for [0=end] : %ld\n",x);
	    if( x!=0 )
	      {
	      historywithBpoints(first, last, x, trans1, trans2, 1, RECfile, PRINTfile);
	      printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	      fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	      }
	    fprintf(stdout, "\n");
	    fprintf(PRINTfile, "\n");
	    }
	   while ( (x!=0) );
	   break;

 case '5': fprintf(stdout,"How many keys ( range ) do you want to search for [0 = end]  : ");
	   fscanf(stdin, "%d",&numberkeys);
	   fprintf(PRINTfile,"How many keys ( range ) do you want to search for [0 = end]  : %d",numberkeys);
	   if(numberkeys > 0)
	    {
	    do
	     {
	     fprintf(stdout, "\nGive me the first of the %d keys whose the history you ask for [0=end] : ",numberkeys);
	     fscanf(stdin, "%ld",&x);
	     fprintf(stdout, "\n");
	     fprintf(PRINTfile,"\nGive me the first of the %d keys whose the history you ask for [0=end] : %ld\n",numberkeys,x);
	     if( x!=0 )
	       {
	       historywithFpoints(first, last, x, trans1, trans2, numberkeys, RECfile, PRINTfile);
	       printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	       fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	       }
	     fprintf(stdout, "\n");
	     fprintf(PRINTfile,"\n");
	     }
	    while ( (x!=0) );
	    } // if(sumkeys > 0)
	   break;

 case '6': fprintf(stdout,"How many keys ( range ) do you want to search for [0 = end]  : ");
	   fscanf(stdin, "%d",&numberkeys);
	   fprintf(PRINTfile,"How many keys ( range ) do you want to search for [0 = end]  : %d",numberkeys);
	   if(numberkeys > 0)
	    {
	    do
	     {
	     fprintf(stdout, "\nGive me the first of the %d keys whose the history you ask for [0=end] : ",numberkeys);
	     fscanf(stdin, "%ld",&x);
	     fprintf(stdout, "\n");
	     fprintf(PRINTfile,"\nGive me the first of the %d keys whose the history you ask for [0=end] : %ld\n",numberkeys,x);
	     if( x!=0 )
	       {
	       historywithBpoints(first, last, x, trans1, trans2, numberkeys, RECfile, PRINTfile);
	       printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	       fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	       }
	     fprintf(stdout, "\n");
	     fprintf(PRINTfile,"\n");
	     }
	    while ( (x!=0) );
	    } // if(sumkeys > 0)
	   break;

  } // switch

}

/*------------------------------------------------------------------------*/

void historywithFpoints( STACKLINK *first, STACKLINK *last, long x,int trans1,
			int trans2, int numberkeys, FILE *RECfile, FILE *PRINTfile)
{
/***********************************************\
| Forward search for the history of key 'x'.    |
| The search starts from the root of the T-tree |
| which corresponds to time point trans1        |
\***********************************************/

 STACKLINK   *headroot;
 int 	     i, found;
 int 	     metritis1 , metritis2,
	     startTrans , endTrans,
	     position , ii , j;
 BTREE       tempnode;

 startTrans = (first->mainroot.transaction<trans1 ? trans1 : first->mainroot.transaction );
 endTrans = (last->mainroot.transaction<trans2 ? last->mainroot.transaction : trans2 );

 headroot=first;

 if(numberkeys==1)
  fprintf(stdout,"Trans ( k )ey   Salary");
 else
  fprintf(stdout,"Trans ( k )ey   Sal.kS.k+1S.k+2S.k+3S.k+4S.k+5S.k+6S.k+7S.k+8S.k+9");
 if(numberkeys==1)
  fprintf(PRINTfile,"Trans ( k )ey   Salary");
 else
  fprintf(PRINTfile,"Trans ( k )ey   Sal.kS.k+1S.k+2S.k+3S.k+4S.k+5S.k+6S.k+7S.k+8S.k+9");

 for(ii=startTrans;ii<=endTrans;ii++)
  {
  for(i=headroot->mainroot.transaction;i<ii;i++)
   headroot=headroot->next;
  if (headroot->mainroot.treepointer!=NULLPOS)
   {
   found=findLEAFnode(x, headroot->mainroot.treepointer, &tempnode, &position, RECfile);
   printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);

   if(numberkeys>1)
    findrestKeysUsingFpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);

   metritis2=( (external(tempnode).endtime<=endTrans) ? external(tempnode).endtime : endTrans );
   metritis1=ii+1;
   for(j=metritis1;j<=metritis2;j++)
    {
    ii++;
    headroot=headroot->next;
    printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);

   if(numberkeys>1)
    findrestKeysUsingFpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);

    }
   while(ii<endTrans)
    {
    ii++;
    headroot=headroot->next;
    if( (external(tempnode).Fpoint)!=NULLPOS )
     {
     tempnode=getBTreeNode(RECfile, external(tempnode).Fpoint);
     bin_search(tempnode, x, &position, &found);
     while( (!found) && (position==(tempnode.contain+1))
     && (external(tempnode).FCpoint!=NULLPOS) && x>maxkeyfor(tempnode) )
      {
      tempnode = getBTreeNode(RECfile, external(tempnode).FCpoint);
      bin_search(tempnode, x, &position, &found);
      }
     printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);

    if(numberkeys>1)
     findrestKeysUsingFpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);

     metritis2=( (external(tempnode).endtime<=endTrans) ? external(tempnode).endtime : endTrans );
     metritis1=ii+1;
     for(j=metritis1;j<=metritis2;j++)
      {
      ii++;
      headroot=headroot->next;
      printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);
     if(numberkeys>1)
      findrestKeysUsingFpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);
      }
     }
    else
     break;
    } // if( (external(tempnode).Fpoint)!=NULLPOS )
   } // if( headroot->mairoot.treepointer!=NULLPOS )
  }
 if(numberkeys==1)
  fprintf(stdout, "\n;\n");
 else
  fprintf(stdout, "\n;\n");

 if(numberkeys==1)
  fprintf(PRINTfile, "\n;\n");
 else
  fprintf(PRINTfile, "\n;\n");

}

/*------------------------------------------------------------------------*/

void historywithBpoints( STACKLINK *first, STACKLINK *last, long x,int trans1,
			int trans2, int numberkeys, FILE *RECfile, FILE *PRINTfile)
{
/************************************************\
| Backward search for the history of key 'x'.    |
| The search starts from the root of the T-tree  |
| which corresponds to time point trans2         |
\************************************************/

 STACKLINK   *headroot;
 int 	     i, found;
 int 	     metritis1 , metritis2,
	     startTrans , endTrans,
	     position , ii , j;
 BTREE       tempnode;

 startTrans = (first->mainroot.transaction<trans1 ? trans1 : first->mainroot.transaction );
 endTrans = (last->mainroot.transaction<trans2 ? last->mainroot.transaction : trans2 );

 headroot=last;

 if(numberkeys==1)
  fprintf(stdout,"Trans ( k )ey   Salary");
 else
  fprintf(stdout,"Trans ( k )ey   Sal.kS.k+1S.k+2S.k+3S.k+4S.k+5S.k+6S.k+7S.k+8S.k+9");

 if(numberkeys==1)
  fprintf(PRINTfile,"Trans ( k )ey   Salary");
 else
  fprintf(PRINTfile,"Trans ( k )ey   Sal.kS.k+1S.k+2S.k+3S.k+4S.k+5S.k+6S.k+7S.k+8S.k+9");

 for(ii=endTrans;ii>=startTrans;ii--)
  {
  for(i=headroot->mainroot.transaction;i>ii;i--)
   headroot = headroot->prev;
  if (headroot->mainroot.treepointer!=NULLPOS)
   {
   found=findLEAFnode(x, headroot->mainroot.treepointer, &tempnode, &position, RECfile);
   printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);

   if(numberkeys>1)
    findrestKeysUsingBpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);

   metritis2=( (tempnode.starttime>=startTrans) ? tempnode.starttime : startTrans );
   metritis1=ii-1;
   for(j=metritis1;j>=metritis2;j--)
    {
    ii--;
    headroot = headroot->prev;
    printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);

   if(numberkeys>1)
    findrestKeysUsingBpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);

    }
   while(ii > startTrans)
    {
    ii--;
    headroot = headroot->prev;
    if( (external(tempnode).Bpoint)!=NULLPOS )
     {
     tempnode=getBTreeNode(RECfile, external(tempnode).Bpoint);
     bin_search(tempnode, x, &position, &found);
     while( (!found) && (position==(tempnode.contain+1))
     && (external(tempnode).BCpoint!=NULLPOS) && x>maxkeyfor(tempnode) )
      {
      tempnode = getBTreeNode(RECfile, external(tempnode).BCpoint);
      bin_search(tempnode, x, &position, &found);
      }
     printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);

     if(numberkeys>1)
      findrestKeysUsingBpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);

     metritis2=( (tempnode.starttime>=startTrans) ? tempnode.starttime : startTrans );
     metritis1=ii-1;
     for(j=metritis1;j>=metritis2;j--)
      {
      ii--;
      headroot = headroot->prev;
      printkey(ii, tempnode, x, position, found, numberkeys, PRINTfile);
      if(numberkeys>1)
       findrestKeysUsingBpoints(tempnode,x,ii,headroot->mainroot.treepointer, numberkeys, RECfile, PRINTfile);
      }
     }
    else
     break;
    } // if( (external(tempnode).Fpoint)!=NULLPOS )
   } // if( headroot->mairoot.treepointer!=NULLPOS )
  }
 if(numberkeys==1)
  fprintf(stdout, "\n;\n");
 else
  fprintf(stdout, "\n;\n");

 if(numberkeys==1)
  fprintf(PRINTfile,"\n;\n");
 else
  fprintf(PRINTfile,"\n;\n");


}

/*------------------------------------------------------------------------*/

void printkey(int ii,BTREE p, long x, int pos, int found, int numberkeys, FILE *PRINTfile)
{
/*********************************************************************\
| It prints the key and the additional secondary information (salary) |     	               		             |
\*********************************************************************/

 if(numberkeys==1)
  {
  fprintf(stdout, "\n͵");
  fprintf(PRINTfile, "\n͵");
  }
 else
  {
  fprintf(stdout, "\n͵");
  fprintf(PRINTfile, "\n͵");
  }

 fprintf(stdout,"\n  No%d ",ii);
 fprintf(stdout, "%12ld",x);
 if( found )
  fprintf(stdout, "%6d",external(p).seckey[pos]);
 else
  fprintf(stdout,"      ");

 fprintf(PRINTfile,"\n  No%d ",ii);
 fprintf(PRINTfile,"%12ld",x);
 if( found )
  fprintf(PRINTfile,"%6d",external(p).seckey[pos]);
 else
  fprintf(PRINTfile,"      ");

}

/*------------------------------------------------------------------------*/

void findrestKeysUsingFpoints( BTREE tempnode, long x, int ii,int rootpos,
			    int sumaskedkeys, FILE *RECfile, FILE *PRINTfile)
{
/************************************************************************\
| When the search conserns a set of keys in a specific                   |
| range [x, y], this routine searches for the keys in the range [x+1, y] |
| by using the leaf node in which we found the key x		         |
\************************************************************************/

long nextx;
int  position,found;

 for(nextx=x+1 ; nextx<x+sumaskedkeys ; nextx++)
  {
  bin_search(tempnode, nextx, &position, &found);
  while( (!found) && (position==(tempnode.contain+1))
  && (external(tempnode).FCpoint!=NULLPOS) && nextx>maxkeyfor(tempnode) )
   {
   tempnode = getBTreeNode(RECfile, external(tempnode).FCpoint);
   if(external(tempnode).endtime<ii)
    {
    found=FALSE;
    break;
    }
   bin_search(tempnode, nextx, &position, &found);
   }
  if( !found  && (position==(tempnode.contain+1))
  && (external(tempnode).FCpoint==NULLPOS) && nextx>maxkeyfor(tempnode) )
   found=findLEAFnode(nextx, rootpos, &tempnode, &position, RECfile);
  else if( !found && (external(tempnode).endtime<ii) )
   found=findLEAFnode(nextx, rootpos, &tempnode, &position, RECfile);

  if( !((nextx-x)%10) )
   {
   fprintf(stdout, "\n      ͵");
   fprintf(PRINTfile, "\n      ͵");
   fprintf(stdout, "\n      %12ld",nextx);
   fprintf(PRINTfile, "\n      %12ld",nextx);
   }
  if( found )
   {
   fprintf(stdout, "%6d",external(tempnode).seckey[position]);
   fprintf(PRINTfile, "%6d",external(tempnode).seckey[position]);
   }
  else
   {
   fprintf(stdout,"      ");
   fprintf(PRINTfile,"      ");
   }

 } // for(nextx=x+1;nextx<x+4;nextx++)
}

/*------------------------------------------------------------------------*/

void findrestKeysUsingBpoints(BTREE tempnode, long x, int ii,int rootpos,
			   int sumaskedkeys, FILE *RECfile, FILE *PRINTfile)
{
/************************************************************************\
| When the search conserns a set of keys in a specific                   |
| range [x, y], this routine searches for the keys in the range [x+1, y] |
| by using the leaf node in which we found the key x		         |
\************************************************************************/

long nextx;
int position,found;

 for(nextx=x+1 ; nextx<x+sumaskedkeys ; nextx++)
  {
  bin_search(tempnode, nextx, &position, &found);
  while( (!found) && (position==(tempnode.contain+1))
  && (external(tempnode).BCpoint!=NULLPOS) && nextx>maxkeyfor(tempnode) )
   {
   tempnode = getBTreeNode(RECfile, external(tempnode).BCpoint);
   if( (tempnode.starttime>ii) || (external(tempnode).endtime<ii) ) // ⤘    ᝜
    {
    found=FALSE;
    break;
    }
   bin_search(tempnode, nextx, &position, &found);
   }
  if( !found  && (position==(tempnode.contain+1))
   && (external(tempnode).BCpoint==NULLPOS) && nextx>maxkeyfor(tempnode) )
   found=findLEAFnode(nextx, rootpos, &tempnode, &position, RECfile);
  else if( !found && (external(tempnode).endtime<ii) )
   found=findLEAFnode(nextx, rootpos, &tempnode, &position, RECfile);

  if( !((nextx-x)%10) )
   {
   fprintf(stdout,"\n      ͵");
   fprintf(PRINTfile, "\n      ͵");
   fprintf(stdout,"\n      %12ld",nextx);
   fprintf(PRINTfile, "\n      %12ld",nextx);
   }
  if( found )
   {
   fprintf(stdout, "%6d",external(tempnode).seckey[position]);
   fprintf(PRINTfile, "%6d",external(tempnode).seckey[position]);
   }
  else
   {
   fprintf(stdout,"      ");
   fprintf(PRINTfile,"      ");
   }
  } // for(nextx=x+1;nextx<x+4;nextx++)
}

/*------------------------------------------------------------------------*/

void TimesliceQuery(char ch, STACKLINK *first, FILE *RECfile, FILE *PRINTfile)
{
/******************************************************\
| This routine is the menu for the timeslice query     |
| depending on the choice of the user between          |
| - pure timeslice queries or                          |
| - range timeslice queries                            |
\******************************************************/

 STACKLINK  *askedtrans;
 long       x1,x2;
 int        rootpos;
 int        trans;

 do
  {
  fprintf(stdout, "\nGive me the transaction [0=end]: ");
  fscanf(stdin, "%d",&trans);
  fprintf(PRINTfile, "\nGive me the transaction [0=end]: %d",trans);
  if (trans!=0)
   {
   askedtrans = findtrans(first, trans);
   rootpos=askedtrans->mainroot.treepointer;
   if( (askedtrans!=NULL) && (rootpos!=NULLPOS) )
    {
    switch( ch )
     {
     case '1': fprintf(stdout, "\nRollback database .\n");
	       fprintf(stdout, "\nTransaction = %d\n\n",trans);
	       fprintf(stdout, "\n  Key#    Salary ");
	       fprintf(stdout, "\n  ----    ------\n");
	       fprintf(PRINTfile,"\nRollback database .\n");
	       fprintf(PRINTfile,"\nTransaction = %d\n\n",trans);
	       fprintf(PRINTfile,"\n  Key#    Salary ");
	       fprintf(PRINTfile,"\n  ----    ------\n");
	       printSliceTree(rootpos, -MAXLONG, MAXLONG, RECfile);
	       printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	       fprintf(PRINTfile,"\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
	       break;

     case '2': do
		{
		fprintf(stdout,"Give me the range of keys to search for [0 0 = end] : ");
		fscanf(stdin, "%ld %ld",&x1,&x2);
		fprintf(PRINTfile,"\nGive me the range of keys to search for [0 0 = end] : %ld %ld\n",x1,x2);
		if( (x1!=0) || (x2!=0) )
		  {
		  fprintf(stdout, "\nRollback database .\n");
		  fprintf(stdout, "\nTransaction = %d\n\n",trans);
		  fprintf(stdout, "\n  Key#    Salary ");
		  fprintf(stdout, "\n  ----    ------\n");
		  fprintf(PRINTfile, "\nRollback database .\n");
		  fprintf(PRINTfile, "\nTransaction = %d\n\n",trans);
		  fprintf(PRINTfile, "\n  Key#    Salary ");
		  fprintf(PRINTfile, "\n  ----    ------\n");
		  printSliceTree(rootpos,x1,x2,RECfile);
		  fprintf(stdout,"\n");
		  fprintf(PRINTfile,"\n");
		  printf("\nWritenNodes = %ld ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
		  fprintf(PRINTfile,"WritenNodes = %d ReadNodes = %ld WritenTrans = %d ReadTrans = %d \n",WritenNodes,ReadNodes,WritenTrans,ReadTrans);
		  }
		fprintf(stdout,"\n");
		fprintf(PRINTfile,"\n");
		}
	       while ( (x1!=0) || (x2!=0) );
	       break;
     }
    }
   else
    {
    fprintf(stdout,"Wrong transaction or transaction does not exists .\n");
    fprintf(PRINTfile,"Wrong transaction or transaction does not exists .\n");
    }
   }
  }
 while ( (trans!=0) );

}

/*------------------------------------------------------------------------*/

void printSliceTree(int ppos, long x1, long x2, FILE *RECfile)
{
/*******************************************************\
| Timeslice query: traverses a specific T-tree and      |
| prints all its keys in the range [x1, x2].            |
| If [x1,x2]=[-infinity, infinity] the query            |
| is called: pure timeslice query.          		|
\*******************************************************/

int   i, found;
int   pos1, pos2;
BTREE p;

if (ppos!=NULLPOS)
 {
 p = getBTreeNode(RECfile, (long)(ppos));
 if ( isLEAF(p) )
  {
  for(i=1;i<=p.contain;i++)
   if( (external(p).key[i]>=x1) && (external(p).key[i]<=x2) )
    fprintf(stdout,"%4ld      %5d\n",external(p).key[i],external(p).seckey[i]);
  }
 else
  {
  bin_search(p,x1,&pos1,&found);
  bin_search(p,x2,&pos2,&found);
  for(i=pos1;i<=pos2;i++)
   printSliceTree(internal(p).p[i],x1,x2,RECfile);
  }
 }
}

/*------------------------------------------------------------------------*/

int findLEAFnode(long x, int rootpos, BTREE *temp, int *pos, FILE *RECfile)
{
/*************************************************\
| This routine finds the leaf node in which       |
| the key 'x' appears or would be appeared        |
| if it had been inserted in the specific T-tree. |
\*************************************************/

BTREE  root;
int    found;        /* boolean */

if(rootpos!=NULLPOS)
 {
 root = getBTreeNode(RECfile, (long)(rootpos));
 if ( !isLEAF(root) )
  {
  bin_search(root, x, pos, &found);
  found=findLEAFnode(x, internal(root).p[*pos], temp, pos, RECfile);
  }
 else
  {
  *temp=root;
  bin_search(root, x, pos, &found);
  }
 }
else // if( rootpos==NULLPOS )
 return(FALSE);

return(found);
}

/*------------------------------------------------------------------------*/

void locate(FILE *RECfile, long rl, long rpos)
{
/********************************************\
| Locate in a file    		             |
\********************************************/

   if( fseek(RECfile, (rl*(rpos-1L)), SEEK_SET)!=0 )
    {
    perror("fseek failed");
    exit(1);
    }
}

/*------------------------------------------------------------------------*/

void writeBTreeNode(FILE *RECfile, BTREE t)
{
/***********************************************\
| Write in the file of the Overlapping B+-trees |
\***********************************************/

   WritenNodes++;
   locate(RECfile, (long)sizeof(BTREE), (long)t.pos);
   fwrite(&t, sizeof(BTREE), 1, RECfile);
}

/*------------------------------------------------------------------------*/

void writeStackNode(FILE *TTTfile, STACK t)
{
/********************************************\
| Write in the file of the TTT structure     |
\********************************************/

   WritenTrans++;
   locate(TTTfile, (long)sizeof(STACK), (long)t.pos);
   fwrite(&t, sizeof(STACK), 1, TTTfile);
}

/*------------------------------------------------------------------------*/

STACK getStackNode(FILE *TTTfile, long rpos)
{
/********************************************\
| Read from the file of the TTT structure    |
\********************************************/

   STACK q;
   ReadTrans++;
   locate(TTTfile, (long)sizeof(STACK), rpos);
   fread(&q, sizeof(STACK), 1, TTTfile);
   return(q);
}

/*------------------------------------------------------------------------*/

BTREE getBTreeNode(FILE *RECfile, long rpos)
{
/************************************************\
| Read from the file of the Overlapping B+-trees |
\************************************************/

   BTREE q;

   ReadNodes++;
   locate(RECfile, (long)sizeof(BTREE), rpos);
   fread(&q, sizeof(BTREE), 1, RECfile);
   return(q);
}

/*------------------------------------------------------------------------*/

void putRootpos(int rootpos, FILE *TTTfile)
{
/********************************************\
| Write the position of the root of the last |
| created T-tree in the TTT file     	     |
\********************************************/

   if( fseek(TTTfile, 0L, SEEK_SET)!=0 )               {
       perror("fseek failed");   exit(1); }
   putw(rootpos, TTTfile);
}

/*------------------------------------------------------------------------*/

int getRootpos(FILE *RECfile)
{
/********************************************\
| Read the position of the root of the last  |
| created T-tree in the TTT file     	     |
\********************************************/

   if( fseek(RECfile, 0L, SEEK_SET)!=0 )               {
       perror("fseek failed");   exit(1); }
   return(getw(RECfile));
}

/*------------------------------------------------------------------------*/

void printtree(int ppos,int *lab, FILE *RECfile, FILE *PRINTfile)
{
/********************************************\
| Traverses a T-tree and prints all its keys |
\********************************************/

int   i,l;
long  k;
BTREE p;


if(ppos!=NULLPOS)
 {
 p = getBTreeNode(RECfile, (long)ppos);
 if ( isLEAF(p) )
  {
  NodesInTree++;
  LeafNodes++;
  *lab=0;
  for(i=1;i<=p.contain;i++)
   {
   PrintedKeys++;
   fprintf(stdout, "%4ld\n",external(p).key[i]);
   }
  }
 else
  {
  NodesInTree++;
  for(i=0;i<=p.contain;i++)
   {
   if (i>0)
    {
    k=internal(p).key[i-1];
    for (l=1;l<=*lab;l++)
     fprintf(stdout, "          ");
    fprintf(stdout,"%4ld\n",k);
    }
   (*lab)--;
   printtree(internal(p).p[i],lab, RECfile, PRINTfile);
   (*lab)++;
   }
  }
 }
}

/*------------------------------------------------------------------------*/

STACKLINK *findtrans(STACKLINK *l,int no)
{
/*****************************************\
| Searches the stacklink list             |
|(main memory list) for a specific record |					     |
\*****************************************/

STACKLINK *temp;

temp = l;
while(temp!=NULL)
 {
 if (temp->mainroot.transaction==no)
   break;
 temp = temp->next;
 }
return( temp );
}

/*------------------------------------------------------------------------*/

REC getnextkey(SL **f,int *flag)
{
/********************************************************\
| Returns the next key for insertion, deletion or update |
\********************************************************/

   SL *q;
   REC rec;

   q = *f;
   if( q==NULL )
    *flag=TRUE;
   else
    *f = (*f)->next;
   rec = q->rec;
   free(q);
return(rec);
}

/*-------------------------------------------------------------------------*/

void makelist(SL **f, FILE *file, int MAXNumKeys)
{
/******************************************\
| Fills the linked list of reocrds         |
| for batch insertion, deletion or update, |
| with MAXNumkeys new records from the     |
| input file 'file'                        |
\******************************************/

REC  a;
char buff[12];
int  eofkeys=FALSE,
     i=1;

 while( (i<=MAXNumKeys) && (!eofkeys) )
  {
  fgets(buff, 15, file);
  if(atol(buff)!=0)
   {
   i++;
   a.key = atol(buff);
   fgets(buff, 15, file);
   a.tag = buff[0];
   inOrder(a, f);
   }
  else
   eofkeys=TRUE;
  }
}

/*-------------------------------------------------------------------------*/

STACK newtreeroot( void )
{
/*******************************************\
| Creates a new node for the stacklink list |
\*******************************************/

STACK        k;

  k.pos=++CountTranspos;
  k.transaction=CountTranspos;
  k.treepointer=NULLPOS;

return(k);
}

/*------------------------------------------------------------------------*/

void getmemForSTACK(STACKLINK **f, STACKLINK **m)
{
/******************************************\
| Inserts a new node in the stacklink list |
\******************************************/

STACKLINK *new;
STACK     a;

if( (new=(STACKLINK *)malloc(sizeof(STACKLINK)))==NULL )
 {
 fprintf(stdout,"Memory allocation failed\n");
 exit(1);
 }

   a = newtreeroot( );
   new->mainroot = a;

   if( (*m)!=NULL )
    {
    new->prev = *m;
    new->next = NULL;
    new->mainroot.treepointer=(*m)->mainroot.treepointer;
    (*m)->next = new;
    *m = new;
    }
   else
    {
    new->prev = NULL;
    new->next = NULL;
    *f = new;
    *m = new;
    }

}

/*------------------------------------------------------------------------*/

void push(REC a, SL **f)
{
/***********************************\
| Inserts a new node in the SL list |
\***********************************/

   SL *q;

   if( (q=(SL *)malloc(sizeof(SL)))==NULL )              {
	printf("Memory allocation failed ... \n");
	getch();
	exit(1);  }

   q->rec = a;
   if( *f==NULL )       {
       q->next = NULL;
       *f = q;          }
   else              {
       q->next = *f;
       *f = q;       }
}

/*-------------------------------------------------------------------------*/

void insAfter(SL **node, REC a)
{
/**********************************************\
| Inserts a new node at the end of the SL list |
\**********************************************/
   SL *q;

   if( *node==NULL )                          {
       printf("  夘 礘\n");
       exit(1);                               }

   if( (q=(SL *)malloc(sizeof(SL)))==NULL )              {
	printf("Memory allocation failed ... \n");
	getch();
	exit(1);  }


   q->rec = a;

   q->next = (*node)->next;
   (*node)->next = q;
}

/*-------------------------------------------------------------------------*/

void inOrder(REC a, SL **f)
{
/*************************************************\
| Inserts a new node in the SL list in a position |
| which depends on the primary key of the record  |
| of the new node.     	               		  |
\*************************************************/

   SL *q, *p;

   q = NULL;
   for(p = *f; p!=NULL && a.key>=p->rec.key; p = p->next)
       q = p;
   if( q==NULL )
       push(a, f);
   else
       insAfter(&q, a);
}

/*------------------------------------------------------------------------*/
void  freeSTACKlist(STACKLINK *f)
{
/********************************************\
| Empty the stacklink list and free the main |
| memory which was allocated    	     |
\********************************************/

 STACKLINK *q;

 while( f!=NULL )
  {
  q = f;
  f = f->next;
  free(q);
  }
}

/*------------------------------------------------------------------------*/

FILE *openRTfile(char file[])
{
/*************************************\
| Open a text file for read and write |
| (if it does not exist, provide run  |
| time error and exit)                |
\*************************************/

FILE *f;

 if( (f=fopen(file,"r+"))==NULL )
    {
     printf("fopen failed\n");
     exit(0);
    }
return(f);
}

/*------------------------------------------------------------------------*/

FILE *openWTfile(char file[])
{
/*************************************\
| Open a text file for write and read |
| (if it does not exist, create it)   |
\*************************************/

FILE *f;

 if( (f=fopen(file,"w+"))==NULL )
    {
     printf("fopen failed\n");
     exit(0);
    }
return(f);
}

/*------------------------------------------------------------------------*/

FILE *openRWfile(char file[])
{
/***************************************\
| Open a binary file for read and write |
| (if it does not exist, provide run    |
| time error and exit)                  |
\***************************************/

FILE *f;

 if( (f=fopen(file, "r+b"))==NULL )
  {
  fprintf(stdout, "Can't open file\n");
  exit(1);
  }
return(f);
}

/*------------------------------------------------------------------------*/

FILE *openWfile(char file[])   /* 櫜   夜 RW */
{
/***************************************\
| Open a binary file for write and read |
| (if it does not exist, create it)     |
\***************************************/

 FILE *f;

 if( (f=fopen(file , "w+b"))==NULL )
  {
  fprintf(stdout,"Can't open file\n");
  exit(1);
  }
 return(f);
}









